package geniusweb.protocol.tournament.allpermutations;

import java.util.Collections;
import java.util.List;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;

import geniusweb.protocol.session.SessionSettings;
import geniusweb.protocol.session.TeamOfPartiesAndProfiles;
import geniusweb.protocol.session.saop.SAOPSettings;
import geniusweb.protocol.session.saop.SaopPartyWithProfile;
import geniusweb.protocol.session.shaop.SHAOPSettings;
import geniusweb.protocol.session.shaop.ShaopTeam;
import geniusweb.protocol.tournament.TournamentProtocol;
import geniusweb.protocol.tournament.TournamentSettings;
import geniusweb.references.PartyWithProfile;
import tudelft.utilities.immutablelist.FixedList;
import tudelft.utilities.immutablelist.Function2;
import tudelft.utilities.immutablelist.ImmutableList;
import tudelft.utilities.immutablelist.MapList;
import tudelft.utilities.immutablelist.MapThreadList;
import tudelft.utilities.immutablelist.Permutations;
import tudelft.utilities.immutablelist.PermutationsOrderedWithoutReturn;
import tudelft.utilities.immutablelist.PermutationsWithReturn;
import tudelft.utilities.immutablelist.PermutationsWithoutReturn;
import tudelft.utilities.immutablelist.Tuple;
import tudelft.utilities.immutablelist.Tuples;
import tudelft.utilities.logging.Reporter;

/**
 * Takes a number of parties and profiles and makes all permutations of them
 * with the given number of parties and profiles. All profiles must be in the
 * same domain.
 * <p>
 * All sessions are based on the given {@link SessionSettings} (can be
 * {@link SAOPSettings} or {@link SHAOPSettings}). The parties for each
 * tournament session are added to this basic sessionsettings. sessionsettings
 * contains basically all the normal run-sessionsettings. This is used as a
 * "template" for all sessions of the tournament. You can put any use any
 * session setting here, and each session will be run according to the protocol
 * you select here.
 * 
 * The participants list usually is empty. The AllPermutationsProtocol adds the
 * the required parties to this list. So if you provide a non-empty list here,
 * then these parties would be present in every session in the tournament.
 * <p>
 * If the tournament is used with a SHAOP sessionsettings, a number of rules are
 * used to generate the COB party from the settings.
 * 
 * The COB party is assumed to be the comparebids-1.0.0 party on the same server
 * as the party you provide. So if you provide http://server/../party1-1.2.3
 * then the cob party will be assumed to be at
 * http://server/../comparebids-1.0.0. The provided parties that you provide are
 * the SHAOP parties (you can also use SAOP parties). The provided profile is
 * also given to the SHAOP party. Be aware that you typically need a parameter
 * like partial=10. The COB party gets the profile as specified, but without the
 * query part. The COB and SHAOP party are thus a fixed pair.
 */
public class AllPermutationsSettings implements TournamentSettings {

	public static final String COB_PARTY = "comparebids-1.2.0";
	private final List<Team> teams;
	private final List<ProfileList> profileslists;
	private final boolean reuseTeams;
	private final int teamsPerSession;
	private final SessionSettings sessionsettings;

	/**
	 * 
	 * @param teams           a list of {@link Team}s. Must contain at least
	 *                        <partiesPerSession> elements. The <teamsize> must
	 *                        match the protocol: (SAOP:1, SHAOP:2)
	 * @param reuseTeams      if true, we use PermutationsWithReturn, if false
	 *                        we use PermutationsWithoutReturn to create the
	 *                        teams.
	 * @param plists          list of available {@link ProfileList}s to be
	 *                        permutated over the teams. Each
	 *                        {@link ProfileList} must contain <teamsize>
	 *                        elements.
	 * @param teamsPerSession number of parties per session, must be at least 2.
	 * @param sesettings      The generic {@link SessionSetting}.
	 *                        {@link SessionSettings#with(PartyWithProfile)}
	 *                        will be used to add the required
	 *                        {@link PartyWithProfile}s
	 */
	@JsonCreator
	public AllPermutationsSettings(@JsonProperty("teams") List<Team> teams,
			@JsonProperty("profileslists") List<ProfileList> plists,
			@JsonProperty("reuseTeams") boolean reuseTeams,
			@JsonProperty("teamsPerSession") int teamsPerSession,
			@JsonProperty("sessionsettings") SessionSettings sesettings) {
		if (teamsPerSession < 2)
			throw new IllegalArgumentException("teamsPerSession must be >=2");
		if (teams == null || teams.size() < teamsPerSession)
			throw new IllegalArgumentException("teams must contain at least "
					+ teamsPerSession + " teams");
		if (plists == null || plists.size() < 2)
			throw new IllegalArgumentException(
					"profileslist must contain at least " + teamsPerSession
							+ " ProfileList's");
		int teamsize;
		if (sesettings instanceof SAOPSettings)
			teamsize = 1;
		else if (sesettings instanceof SHAOPSettings)
			teamsize = 2;
		else
			throw new IllegalArgumentException(
					"Unsupported protocol " + sesettings.getClass());

		for (Team team : teams) {
			if (team.getParties().size() != teamsize) {
				throw new IllegalArgumentException("All teams should have size "
						+ teamsize + " but found " + team.getParties().size());
			}
		}
		for (ProfileList profile : plists) {
			if (profile.getProfiles().size() != teamsize) {
				throw new IllegalArgumentException(
						"All profiles should have size equal to the team size "
								+ teamsize + " but found "
								+ profile.getProfiles().size());

			}
		}
		this.teamsPerSession = teamsPerSession;
		this.reuseTeams = reuseTeams;
		this.teams = teams;
		this.profileslists = plists;
		this.sessionsettings = sesettings;
	}

	@Override
	public Double getMaxRunTime() {
		return permutations().size().doubleValue()
				* (sessionsettings.getMaxRunTime() + 2d);
	}

	@Override
	public TournamentProtocol getProtocol(Reporter logger) {
		AllPermutationsState state = new AllPermutationsState(this,
				permutations(), Collections.emptyList());
		return new AllPermutationsProtocol(state, logger);
	}

	/**
	 * @return {@link ImmutableList} containing all {@link SessionSettings} for
	 *         all Sessions in the correct order.
	 */
	public ImmutableList<SessionSettings> permutations() {
		ImmutableList<ImmutableList<TeamOfPartiesAndProfiles>> partylistlist;
		ImmutableList<Team> partieslist = new FixedList<Team>(teams);
		ImmutableList<ProfileList> profileslist = new FixedList<ProfileList>(
				profileslists);
		partylistlist = getParticipants(partieslist, profileslist,
				teamsPerSession, reuseTeams);
		return new MapList<>(partyproflist -> createSetting(partyproflist),
				partylistlist);
	}

	@Override
	public String toString() {
		return "AllPermutationsSettings[" + teams + "," + reuseTeams + ","
				+ profileslists + "," + teamsPerSession + "," + sessionsettings
				+ "]";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((teams == null) ? 0 : teams.hashCode());
		result = prime * result + teamsPerSession;
		result = prime * result
				+ ((profileslists == null) ? 0 : profileslists.hashCode());
		result = prime * result + (reuseTeams ? 1231 : 1237);
		result = prime * result
				+ ((sessionsettings == null) ? 0 : sessionsettings.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AllPermutationsSettings other = (AllPermutationsSettings) obj;
		if (teams == null) {
			if (other.teams != null)
				return false;
		} else if (!teams.equals(other.teams))
			return false;
		if (teamsPerSession != other.teamsPerSession)
			return false;
		if (profileslists == null) {
			if (other.profileslists != null)
				return false;
		} else if (!profileslists.equals(other.profileslists))
			return false;
		if (reuseTeams != other.reuseTeams)
			return false;
		if (sessionsettings == null) {
			if (other.sessionsettings != null)
				return false;
		} else if (!sessionsettings.equals(other.sessionsettings))
			return false;
		return true;
	}

	/**
	 * 
	 * @param partyproflist a list of {@link PartyWithProfile} settings
	 * @return {@link SessionSettings}
	 */
	private SessionSettings createSetting(
			ImmutableList<TeamOfPartiesAndProfiles> partyproflist) {
		SessionSettings settings = sessionsettings;
		for (TeamOfPartiesAndProfiles partyprof : partyproflist) {
			settings = settings.with(partyprof);
		}
		return settings;
	}

	/**
	 * 
	 * @param parties
	 * @param profiles
	 * @param n                    number of items to draw. Assumed >= 1.
	 * @param drawPartyWithPutback if parties can be drawn multiple times
	 * @return list of sessionsets where each sessionset contains settings for a
	 *         session: a list with {@link TeamOfPartiesAndProfiles}s. The
	 *         sessionsets are made by making all combinations of parties and
	 *         profiles. Profiles are drawn with replace.
	 */
	private ImmutableList<ImmutableList<TeamOfPartiesAndProfiles>> getParticipants(
			ImmutableList<Team> parties, ImmutableList<ProfileList> profiles,
			int n, boolean drawPartyWithPutback) {

		Permutations<Team> partiesPermutations;
		if (drawPartyWithPutback) {
			partiesPermutations = new PermutationsWithReturn<>(parties, n);
		} else {
			partiesPermutations = new PermutationsWithoutReturn<>(parties, n);
		}

		Permutations<ProfileList> profilesPermutations = new PermutationsOrderedWithoutReturn<>(
				profiles, n);

		// each tuple contains session info: a list of Team and a list of
		// ProfileList.
		Tuples<ImmutableList<Team>, ImmutableList<ProfileList>> tuples = new Tuples<>(
				partiesPermutations, profilesPermutations);

		return new MapList<Tuple<ImmutableList<Team>, ImmutableList<ProfileList>>, ImmutableList<TeamOfPartiesAndProfiles>>(
				tuple -> teamlist(tuple), tuples);
	}

	/**
	 * 
	 * @param tuple a tuple with (1) ImmutableList<Team> (2)
	 *              ImmutableList<Profiles>
	 * @return list of TeamOfPartiesAndProfiles, each picked from the two lists
	 *         in order.
	 */
	private ImmutableList<TeamOfPartiesAndProfiles> teamlist(
			Tuple<ImmutableList<Team>, ImmutableList<ProfileList>> tuple) {
		Function2<Team, ProfileList, TeamOfPartiesAndProfiles> makeparty = new Function2<Team, ProfileList, TeamOfPartiesAndProfiles>() {
			@Override
			public TeamOfPartiesAndProfiles apply(Team team,
					ProfileList profilelist) {
				if (sessionsettings instanceof SAOPSettings)
					return new SaopPartyWithProfile(team.getParties().get(0),
							profilelist.getProfiles().get(0));
				else if (sessionsettings instanceof SHAOPSettings) {
					return new ShaopTeam(
							new PartyWithProfile(team.getParties().get(0),
									profilelist.getProfiles().get(0)),
							new PartyWithProfile(team.getParties().get(1),
									profilelist.getProfiles().get(1)));
				} else
					throw new IllegalArgumentException(
							"Unknown/unsupported protocol "
									+ sessionsettings.getClass());
			}

		};
		return new MapThreadList<TeamOfPartiesAndProfiles, Team, ProfileList>(
				makeparty, tuple.get1(), tuple.get2());
	}

//	/**
//	 * Extract the raw URL, without the part after the question mark
//	 * 
//	 * @param party   a party reference (for another party)
//	 * @param profile a profile that may have a filter like "?partial=XX"
//	 * @return profile without the filter and with a cob party on the same
//	 *         machine. IT IS ASSUMED that there will the default
//	 *         "comparebids-1.0.0" party on the same machine as where the other
//	 *         party is
//	 */
//	protected static PartyWithProfile makeCob(PartyRef party,
//			ProfileRef profile) {
//		try {
//			URI profileuri = profile.getURI();
//			URI partyuri = party.getURI();
//			ProfileRef cobprof = new ProfileRef(profileuri.getScheme() + "://"
//					+ profileuri.getAuthority() + profileuri.getPath());
//			String partypath = partyuri.getPath();
//			partypath = partypath.substring(0, partypath.lastIndexOf('/'));
//			PartyWithParameters cobparty = new PartyWithParameters(new PartyRef(
//					partyuri.getScheme() + "://" + partyuri.getAuthority()
//							+ partypath + "/" + COB_PARTY),
//					new Parameters());
//			return new PartyWithProfile(cobparty, cobprof);
//
//		} catch (URISyntaxException e) {
//			throw new IllegalArgumentException(
//					"Failed making cob party with profile " + profile, e);
//		}
//	}

}
